/*
 * Copyright (C) 2009 by Eric Lambert <Eric.Lambert@sun.com>
 * Use and distribution licensed under the
 * GNU Lesser General Public License (LGPL) version 2.1.
 * See the COPYING file in the parent directory for full text.
 */

package org.gearman.common;


import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;
import java.util.logging.Level;
import java.util.logging.Logger;


public class GearmanNIOJobServerConnection implements GearmanJobServerConnection{

    private InetSocketAddress remote;
    private SocketChannel serverConnection = null;
    private Selector selector = null;
    private SelectionKey selectorKey = null;
    private static final Logger LOG = Logger.getLogger(
            Constants.GEARMAN_CLIENT_LOGGER_NAME); //TODO change this
    private ByteBuffer bytesReceived;
    private ByteBuffer bytesToSend;

    public GearmanNIOJobServerConnection(String hostname)
            throws IllegalArgumentException {
        this(hostname,Constants.GEARMAN_DEFAULT_TCP_PORT);
    }

    public GearmanNIOJobServerConnection(String hostname, int port)
            throws IllegalArgumentException {
        this(new InetSocketAddress(hostname, port));
    }

    public GearmanNIOJobServerConnection(InetSocketAddress remote)
            throws IllegalArgumentException {
        if (remote == null) {
            throw new IllegalArgumentException("Remote can not be null");
        }
        this.remote = remote;
        bytesReceived = ByteBuffer.allocate(Constants.GEARMAN_DEFAULT_SOCKET_RECV_SIZE);
        bytesToSend = ByteBuffer.allocate(Constants.GEARMAN_DEFAULT_SOCKET_RECV_SIZE);
    }

    public void open() throws IOException{
        if (isOpen()) {
            throw new IllegalStateException("A session can not be " +
                    "initialized twice");
        }
        try {
            serverConnection = SocketChannel.open(remote);
            serverConnection.configureBlocking(false);
            serverConnection.finishConnect();
            selector = Selector.open();
            selectorKey = serverConnection.register(selector,
                    SelectionKey.OP_WRITE | SelectionKey.OP_READ);
        } catch (IOException ioe) {
            LOG.log(Level.WARNING, "Received IOException while attempting to" +
                    " initialize session " + this +
                    ". Shuting down session", ioe);
            if (serverConnection != null && serverConnection.isOpen()) {
                if (selector != null && selector.isOpen()) {
                    try {
                        selector.close();
                    } catch (IOException selioe) {
                        LOG.log(Level.WARNING, "Received IOException while" +
                                " attempting to close selector.", selioe);
                    }
                }
                try {
                    serverConnection.close();
                } catch (IOException closeioe) {
                    LOG.log(Level.WARNING, "Received IOException while" +
                            " attempting to close connection to server. " +
                            "Giving up!", closeioe);
                }
            }
            throw new IOException();
        }
    }

    public void close() {
        if (!isOpen()) {
            throw new IllegalStateException("Can not close a session that " +
                    "has not been initialized");
        }
        LOG.log(Level.FINE, "Session " + this + " is being closed.");
        selectorKey.cancel();
        try {
            selector.close();
        } catch (IOException ioe) {
            LOG.log(Level.WARNING, "Received IOException while attempting to " +
                    "close selector attached to session " + this, ioe);
        } finally {
            try {
                serverConnection.close();
            } catch (IOException cioe) {
                LOG.log(Level.WARNING, "Received IOException while attempting" +
                        " to close connection for session " + this, cioe);
            }
            serverConnection = null;
        }
        LOG.log(Level.FINE, "Session " + this + " has successfully closed.");
    }

    public void write(GearmanPacket request) throws IOException {

        if (request == null && bytesToSend.position() == 0) {
            return;
        }
        if (request != null) {
            int ps = request.getData().length + Constants.GEARMAN_PACKET_HEADER_SIZE;
            if (bytesToSend.remaining() < ps) {
                //TODO allocate more
                ByteBuffer bb = ByteBuffer.allocate(bytesToSend.capacity() +
                        (ps * 10));
                bb.put(bytesToSend);
                bytesToSend = bb;
            }
            byte[] bytes = request.toBytes();
            ByteBuffer bb = ByteBuffer.allocate(bytes.length);
            bb.put(bytes);
            bb.rewind();
            bytesToSend.put(bb);
        }
        selector.selectNow();
        if (selectorKey.isWritable()) {
            bytesToSend.limit(bytesToSend.position());
            bytesToSend.rewind();
            serverConnection.write(bytesToSend);
            bytesToSend.compact();
        } else {
            System.out.println("FAILED");
        }
    }

    public GearmanPacket read() throws IOException {
        GearmanPacket returnPacket = null;
        selector.selectNow();
        if (selectorKey.isReadable()) {
            int bytesRead = serverConnection.read(bytesReceived);
            if (bytesRead >= 0) {
                LOG.log(Level.FINER, "Session " + this + " has read " +
                        bytesRead + " bytes from its job server. Buffer has " +
                        bytesReceived.remaining());
            } else {
                //TODO do something smarter here
                throw new IOException("Connection to job server severed");
            }
        }
        if (bufferContainsCompletePacket(bytesReceived)) {
            byte[] pb = new byte[getSizeOfPacket(bytesReceived)];
            bytesReceived.limit(bytesReceived.position());
            bytesReceived.rewind();
            bytesReceived.get(pb);
            bytesReceived.compact();
            returnPacket = new Packet(new BufferedInputStream(
                    new ByteArrayInputStream(pb)));
        }
        return returnPacket;
    }

    public SelectionKey registerSelector(Selector s, int mask)
            throws IOException{
        return serverConnection.register(s, mask);

    }

    public boolean canRead () {
        if (!selector.isOpen()) {
            //LOG.log(Level.WARNING,"Selector is not open, will return false");
            return false;
        }
        try {
            selector.selectNow();
        } catch (IOException ioe) {
            LOG.log(Level.WARNING,"Failed to select on connection " + this,ioe);
        }
        //LOG.log(Level.WARNING,"Selector is readable == " + selectorKey.isReadable() + " and bytes in buffer = " + bytesReceived.position());
        return (selectorKey.isReadable() ||
                bufferContainsCompletePacket(bytesReceived));
    }

    public boolean canWrite () {
        if (!selector.isOpen()) {
            LOG.log(Level.WARNING,"NIO_ Selector is not open, will return false");
            return false;
        }
        try {
            selector.selectNow();
        } catch (IOException ioe) {
            LOG.log(Level.WARNING,"NIO_ Failed to select on connection " + this,ioe);
        }
        LOG.log(Level.WARNING,"Selector is writable == " + selectorKey.isWritable() + " and bytes in buffer = " + bytesToSend.position());
        return (selectorKey.isWritable());
    }

    public boolean hasBufferedWriteData() {
        return bytesToSend.position() >  0;
    }

    public Selector getSelector() {
        return selector;
    }

    public boolean isOpen() {
        return (serverConnection != null && serverConnection.isConnected());
    }

    @Override
    public boolean equals(Object that) {
        if (that == null) {
            return false;
        }

        if (this == that) {
            return true;
        }

        if (! (that instanceof GearmanNIOJobServerConnection )) {
            return false;
        }

        InetSocketAddress thatRemote = ((GearmanNIOJobServerConnection)that).remote;

        return this.remote.equals(thatRemote);
    }

    // When you override equals you should override hashcode as well, since
    // two equal objects should have the same hashcode
    @Override
    public int hashCode() {
        return this.remote != null ? this.remote.hashCode() : 0;
    }



    private boolean bufferContainsCompletePacket(ByteBuffer b) {
        if (b.position() < Constants.GEARMAN_PACKET_HEADER_SIZE ) {
            return false;
        }
        return b.position() >= getSizeOfPacket(b) ? true : false;
    }

    // DO NOT CALL UNLESS YOU ARE SURE THAT BYTEBUFFER HAS AT LEAST
    // GEARMAN_PACKET_HEADER_SIZE BYTES!
    private int getSizeOfPacket(ByteBuffer buffer) {
        int originalPosition = buffer.position();
        byte[] header =new byte[Constants.GEARMAN_PACKET_HEADER_SIZE];
        buffer.rewind();
        buffer.get(header);
        buffer.position(originalPosition);
        PacketHeader ph = new PacketHeader(header);
        return ph.getDataLength() + Constants.GEARMAN_PACKET_HEADER_SIZE;
    }
}
