/*
 * Copyright (C) 2009 by Eric Lambert <Eric.Lambert@sun.com>
 * Use and distribution licensed under the
 * GNU Lesser General Public License (LGPL) version 2.1.
 * See the COPYING file in the parent directory for full text.
 */
package org.gearman.worker;


import org.gearman.common.GearmanSessionEvent;
import org.gearman.common.PacketMagic;
import org.gearman.common.PacketType;
import java.io.IOException;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.concurrent.Executors;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.gearman.common.Constants;
import org.gearman.common.GearmanException;
import org.gearman.common.GearmanPacket;
import org.gearman.common.GearmanJobServerConnection;
import org.gearman.common.GearmanJobServerSession;
import org.gearman.common.GearmanTask;
import org.gearman.common.GearmanServerResponseHandler;
import org.gearman.common.GearmanSessionEventHandler;
import org.gearman.common.Packet;
import org.gearman.util.ByteUtils;
import org.gearman.util.IORuntimeException;

public class GearmanWorkerImpl implements GearmanWorker, GearmanSessionEventHandler {

    public void handleSessionEvent(GearmanSessionEvent event)
            throws IllegalArgumentException, IllegalStateException {
        GearmanPacket p = event.getPacket();
        GearmanJobServerSession s = event.getSession();
        PacketType t = p.getPacketType();
        switch (t) {
            case JOB_ASSIGN:
                taskMap.remove(s);
                addNewJob(event);
                break;
            case JOB_ASSIGN_UNIQ:
                taskMap.remove(s);
                addNewJob(event);
                break;
            case NOOP:
                taskMap.remove(s);
                break;
            case NO_JOB:
                taskMap.put(s, new GearmanTask(new GrabJobEventHandler(s, this), new Packet(PacketMagic.REQ, PacketType.PRE_SLEEP, new byte[0])));
                break;
            case ECHO_RES:
                break;
            case OPTION_RES:
                break;
            case ERROR:
                s.closeSession();
                break;
        }
    }

    private void addNewJob(GearmanSessionEvent event) {
        byte[] handle, data, functionNameBytes;
        GearmanPacket p = event.getPacket();
        GearmanJobServerSession sess = event.getSession();
        String functionName;
        handle = p.getDataComponentValue(
                GearmanPacket.DataComponentName.JOB_HANDLE);
        functionNameBytes = p.getDataComponentValue(
                GearmanPacket.DataComponentName.FUNCTION_NAME);
        data = p.getDataComponentValue(
                GearmanPacket.DataComponentName.DATA);
        functionName = ByteUtils.fromUTF8Bytes(functionNameBytes);
        FunctionDefinition def = functionMap.get(functionName);
        if (def == null) {
            GearmanTask gsr = new GearmanTask(
                    new Packet(PacketMagic.REQ,
                    PacketType.WORK_FAIL, handle));
            sess.submitTask(gsr);
        }
        GearmanFunction function = def.getFactory().getFunction();
        function.setData(data);
        function.setJobHandle(handle);
        function.registerEventListener(sess);
        functionList.add(function);
    }

    public enum State {

        IDLE, RUNNING, SHUTTINGDOWN
    }

    private static final String DESCRIPION_PREFIX = "GearmanWorker";
    private final String DESCRIPTION;
    private ArrayList<GearmanJobServerSession> sessionList = null;
    private ArrayList<GearmanFunction> functionList = null;
    private Selector ioAvailable = null;
    private static final Logger LOG = Logger.getLogger(
            Constants.GEARMAN_CLIENT_LOGGER_NAME);
    private String id;
    private HashMap<String, FunctionDefinition> functionMap;
    private State state;
    private ExecutorService executorService;
    private HashMap<GearmanJobServerSession,GearmanTask> taskMap = null;
    private HashMap<GearmanJobServerSession,Collection<GearmanPacket>> newJobs = null;

    class GrabJobEventHandler implements GearmanServerResponseHandler {

        private GearmanJobServerSession session;
        private boolean isDone = false;
        private GearmanWorkerImpl worker;

        GrabJobEventHandler(GearmanJobServerSession session, GearmanWorkerImpl worker) {
            super();
            this.session = session;
            this.worker = worker;
        }

        public void handleEvent(GearmanPacket event) throws GearmanException {
            handleSessionEvent(new GearmanSessionEvent(event, session));
            isDone = true;
        }

        public boolean isDone() {
            return isDone;
        }
    }

    class GearmanFunctionNameFactory implements GearmanFunctionFactory {

        private String className = null;

        GearmanFunctionNameFactory(String className) {
            this.className = className;
        }

        public String getFunctionName() {
            return className;
        }

        public GearmanFunction getFunction() {
            GearmanFunction f = null;
            try {
                Class c = Class.forName(className);
                Object o = c.newInstance();
                if (o instanceof GearmanFunction) {
                    f = (GearmanFunction) o;
                }
            } catch (Exception e) {
                LOG.log(Level.WARNING, "Unable to create instance of " +
                        "Function: " + className, e);
            }
            return f;
        }
    }

    class FunctionDefinition {

        private long timeout = 0;
        private GearmanFunctionFactory factory = null;

        FunctionDefinition(long timeout, GearmanFunctionFactory factory) {
            this.timeout = timeout;
            this.factory = factory;
        }

        long getTimeout() {
            return timeout;
        }

        GearmanFunctionFactory getFactory() {
            return factory;
        }
    }

    public GearmanWorkerImpl() {
        this(Executors.newSingleThreadExecutor());
    }

    public GearmanWorkerImpl(ExecutorService executorService) {
        DESCRIPTION = DESCRIPION_PREFIX + ":" + Thread.currentThread().getId();
        sessionList = new ArrayList<GearmanJobServerSession>();
        functionList = new ArrayList<GearmanFunction>();
        id = DESCRIPTION;
        functionMap = new HashMap<String, FunctionDefinition>();
        state = State.IDLE;
        this.executorService = executorService;
        taskMap = new HashMap<GearmanJobServerSession, GearmanTask>();
        newJobs = new HashMap<GearmanJobServerSession, Collection<GearmanPacket>>();
    }

    public String toString() {
        return id;
    }

    public void work() {
        if (!state.equals(State.IDLE)) {
            throw new IllegalStateException("Can not call work while worker " +
                    "is running or shutting down");
        }

        state = State.RUNNING;
        while (isRunning()) {

            for (GearmanJobServerSession sess : sessionList) {
                if (!sess.isInitialized()) {
                    continue;
                }
                try {
                    GearmanTask sessTask = taskMap.get(sess);
                    if (sessTask == null) {
                        sessTask = new GearmanTask(
                                new GrabJobEventHandler(sess, this),
                                new Packet(PacketMagic.REQ,
                                PacketType.GRAB_JOB, new byte[0]));
                        taskMap.put(sess, sessTask);
                    }
                    if (sessTask.getState().equals(GearmanTask.State.NEW)) {
                        sess.submitTask(sessTask);
                    }
                    sess.driveSessionIO();
                } catch (IOException ioe) {
                    LOG.log(Level.WARNING, "Received IOException while driving" +
                            " IO on session " + sess, ioe);
                    sess.closeSession();
                    continue;
                }
            }

            for (GearmanFunction fun : functionList) {
                Future<GearmanPacket> gp = executorService.submit(fun);
                FunctionDefinition def = functionMap.get(fun.getName());
                try {
                    if (def.getTimeout() > 0) {
                        gp.get(functionMap.get(fun.getName()).getTimeout(),
                                TimeUnit.MILLISECONDS);
                    } else {
                        gp.get();
                    }
                } catch (Exception e) {
                    fun.fireEvent(new Packet(PacketMagic.REQ,
                            PacketType.WORK_EXCEPTION,
                            Packet.generatePacketData(fun.getJobHandle(),
                            e.getMessage().getBytes())));
                    fun.fireEvent(new Packet(PacketMagic.REQ,
                            PacketType.WORK_FAIL, fun.getJobHandle()));
                }
            }
            functionList.clear();

            boolean shouldSleep = true;
            for (GearmanJobServerSession sess : sessionList) {
                GearmanTask sessTask = taskMap.get(sess);
                if (sessTask == null || !(sessTask.getRequestPacket().getPacketType().equals(PacketType.PRE_SLEEP))) {
                    shouldSleep = false;
                }
            }
            if (shouldSleep) {
                try {
                    Thread.currentThread().sleep(250);
                } catch (InterruptedException ignore) {
                }
            }
        }

        shutDownWorker(true);
    }

    public void addServer(GearmanJobServerConnection conn)
            throws IllegalArgumentException, IllegalStateException {

        //this is a sub-optimal way to look for dups, but addJobServer
        //ops should be infrequent enough that this should be a big penalty
        Iterator<GearmanJobServerSession> iter = sessionList.iterator();
        while (iter.hasNext()) {
            if (iter.next().getConnection().equals(conn)) {
                return;
            }
        }

        GearmanJobServerSession session =
                new GearmanJobServerSession(conn);
        sessionList.add(session);
        if (ioAvailable == null) {
            try {
                ioAvailable = Selector.open();
            } catch (IOException ioe) {
                throw new IORuntimeException(ioe);
            }
        }
        try {
            session.initSession(ioAvailable,
                    SelectionKey.OP_READ | SelectionKey.OP_WRITE, this);
        } catch (IOException ioe) {
            throw new IORuntimeException(ioe);
        }

        GearmanPacket p = new Packet(PacketMagic.REQ, PacketType.SET_CLIENT_ID,
                ByteUtils.toUTF8Bytes(id));
        session.submitTask(new GearmanTask(p));

        for (FunctionDefinition def : functionMap.values()) {
            p = generateCanDoPacket(def);
            session.submitTask(new GearmanTask(p));
        }

        if (isRunning()) {
            p = new Packet(PacketMagic.REQ, PacketType.GRAB_JOB, new byte[0]);
            GearmanTask gsr = new GearmanTask(
                    new GrabJobEventHandler(session,this), p);
            session.submitTask(gsr);
        }
        LOG.log(Level.FINE, "Added server " + conn + " to worker " + this);
    }

    public boolean hasServer(GearmanJobServerConnection conn) {
        boolean foundIt = false;
        for (GearmanJobServerSession sess : sessionList) {
            if (sess.getConnection().equals(conn)) {
                foundIt = true;
            }
        }
        return foundIt;
    }

    public String echo(String text, GearmanJobServerConnection conn) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void registerFunction(String function, long timeout) {
        registerFunctionFactory(new GearmanFunctionNameFactory(function), timeout);
    }

    public void registerFunction(String function) {
        registerFunction(function, 0);
    }

    public void registerFunction(Class<? extends GearmanFunction> function) {
        registerFunction(function, 0);
    }

    public void registerFunction(Class<? extends GearmanFunction> function,
            long timeout) {
        registerFunctionFactory(new GearmanFunctionNameFactory(
                function.getCanonicalName()), timeout);
    }

    public void registerFunctionFactory(GearmanFunctionFactory factory) {
        registerFunctionFactory(factory, 0);
    }

    public void registerFunctionFactory(GearmanFunctionFactory factory,
            long timeout) {
        if (functionMap.containsKey(factory.getFunctionName())) {
            return;
        }
        FunctionDefinition def = new FunctionDefinition(timeout, factory);
        functionMap.put(factory.getFunctionName(), def);
        sendToAll(generateCanDoPacket(def));
    }

    public Set<String> getRegisteredFunctions() {
        HashSet<String> functions = new HashSet<String>();
        for (FunctionDefinition def : functionMap.values()) {
            functions.add(def.factory.getFunctionName());
        }
        return functions;
    }

    public void setWorkerID(String id) throws IllegalArgumentException {
        if (id == null) {
            throw new IllegalArgumentException("Worker ID may not be null");
        }
        this.id = id;
        sendToAll(new Packet(PacketMagic.REQ, PacketType.SET_CLIENT_ID,
                ByteUtils.toUTF8Bytes(id)));
    }

    public void setWorkerID(String id, GearmanJobServerConnection conn) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public String getWorkerID() {
        return id;
    }

    public String getWorkerID(GearmanJobServerConnection conn) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void unregisterFunction(String functionName) {
        functionMap.remove(functionName);
        sendToAll(new Packet(PacketMagic.REQ, PacketType.CANT_DO,
                ByteUtils.toUTF8Bytes(functionName)));
    }

    public void unregisterAll() {
        functionMap.clear();
        sendToAll(new Packet(PacketMagic.REQ, PacketType.RESET_ABILITIES,
                new byte[0]));
    }

    public void stop() {
        state = State.SHUTTINGDOWN;
    }

    public List<Exception> shutdown() {
        return shutDownWorker(false);
    }

    public boolean isRunning() {
        return state.equals(State.RUNNING);
    }


    private GearmanPacket generateCanDoPacket(FunctionDefinition def) {
        PacketType pt = PacketType.CAN_DO;
        byte[] data = null;
        byte[] name = ByteUtils.toUTF8Bytes(def.getFactory().getFunctionName());
        long timeout = def.getTimeout();

        if (timeout > 0) {
            pt = PacketType.CAN_DO_TIMEOUT;
            byte[] to = ByteUtils.toUTF8Bytes(String.valueOf(timeout));
            data = new byte[name.length + to.length + 1];
            System.arraycopy(name, 0, data, 0, name.length);
            data[name.length] = ByteUtils.NULL;
            System.arraycopy(to, 0, data, name.length + 1, to.length);
        } else {
            data = name;
        }
        return new Packet(PacketMagic.REQ, pt, data);
    }

    private void sendToAll(GearmanPacket p) {
        sendToAll(null, p);
    }

    private void sendToAll(GearmanServerResponseHandler handler, GearmanPacket p) {
        GearmanTask gsr = null;
        if (handler == null) {
            gsr = new GearmanTask(p);
        } else {
            gsr = new GearmanTask(handler, p);
        }
        for (GearmanJobServerSession sess : sessionList) {
            sess.submitTask(gsr);
        }

    }

    /*
     * For the time being this will always return an empty list of
     * exceptions because closeSession does not throw an exception
     */
    private List<Exception> shutDownWorker(boolean completeTasks) {

        ArrayList<Exception> exceptions = new ArrayList<Exception>();

        // This gives any jobs in flight a chance to complete
        if (completeTasks) {
            executorService.shutdown();
        } else {
            executorService.shutdownNow();
        }

        for (GearmanJobServerSession sess : sessionList) {
            sess.closeSession();
        }
        state = State.IDLE;

        return exceptions;
    }
}
